# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def sanitize_identifier(identifier):
    # type: (str) -> str
    """
    Get santized identifier name for identifiers which happen to be python
    keywords.
    """

    keywords = (
        "and",
        "as",
        "assert",
        "break",
        "class",
        "continue",
        "def",
        "del",
        "elif",
        "else",
        "except",
        "exec",
        "finally",
        "for",
        "from",
        "global",
        "if",
        "import",
        "in",
        "is",
        "lambda",
        "not",
        "or",
        "pass",
        "print",
        "raise",
        "return",
        "try",
        "while",
        "with",
        "yield",
    )

    return "{}_".format(identifier) if identifier in keywords else identifier



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Type, Any, Union
from enum import Enum

from .case import to_snake_case
from .sanitize import sanitize_identifier


class ConjureType(object):
    pass


DecodableType = Union[
    int, float, bool, str, ConjureType, List[Any], Dict[Any, Any]
]


class ListType(ConjureType):
    item_type = None  # type: Type[DecodableType]

    def __init__(self, item_type):
        # type: (Type[DecodableType]) -> None
        self.item_type = item_type


class DictType(ConjureType):
    key_type = None  # type: Type[DecodableType]
    value_type = None  # type: Type[DecodableType]

    def __init__(self, key_type, value_type):
        # type: (Type[DecodableType], Type[DecodableType]) -> None
        self.key_type = key_type
        self.value_type = value_type


class OptionalType(ConjureType):
    item_type = None  # type: Type[DecodableType]

    def __init__(self, item_type):
        # type: (Type[DecodableType]) -> None
        self.item_type = item_type


class BinaryType(ConjureType):
    pass


class ConjureEnumType(ConjureType, Enum):

    # override __format__ of ConjureEnumType to prevent issue with default
    # Enum behaviour due to lack of __format__ method on ConjureType
    def __format__(self, format_spec):
        return self.__str__()

    def __repr__(self):
        return "{}.{}".format(self.__class__.__name__, self.value)


class ConjureBeanType(ConjureType):

    @classmethod
    def _fields(cls):
        # type: () -> Dict[str, ConjureFieldDefinition]
        """_fields is a mapping from constructor argument
        name to the field definition"""
        return {}

    def __eq__(self, other):
        # type: (Any) -> bool
        if not isinstance(other, self.__class__):
            return False

        for attr in self._fields():
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        # type: (Any) -> bool
        return not self == other

    def __repr__(self):
        # type: () -> str
        fields = [
            "{}={}".format(attr, repr(getattr(self, attr)))
            for attr, field_def in self._fields().items()
        ]
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))


ConjureTypeType = Union[ConjureType, Type[DecodableType]]


class ConjureUnionType(ConjureType):
    _type = None  # type: str

    @property
    def type(self):
        # type: () -> str
        """the member name present in this union"""
        return self._type

    @classmethod
    def _options(cls):
        # type: () -> Dict[str, ConjureFieldDefinition]
        """_options defines a mapping from each member in the union
        to the field definition for that type"""
        return {}

    def __eq__(self, other):
        # type: (Any) -> bool
        if not isinstance(other, self.__class__):
            return False

        assert isinstance(other, ConjureUnionType)

        pythonic_sanitized_identifier = \
            sanitize_identifier(to_snake_case(self.type))

        return other.type == self.type and \
            getattr(self, pythonic_sanitized_identifier) == \
            getattr(other, pythonic_sanitized_identifier)

    def __ne__(self, other):
        # type: (Any) -> bool
        return not self == other

    def __repr__(self):
        # type: () -> str
        fields = [
            "{}={}".format(attr, repr(getattr(self, attr)))
            for attr, field_def in self._options().items()
            if getattr(self, attr) is not None
        ]
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))


class ConjureFieldDefinition(object):
    identifier = None  # type: str
    field_type = None  # type: ConjureTypeType

    def __init__(self, identifier, field_type):
        # type: (str, ConjureTypeType) -> None
        self.identifier = identifier
        self.field_type = field_type



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Any


class SslConfiguration(object):
    trust_store_path = None  # type: str

    def __init__(self, trust_store_path):
        # type: (str) -> None
        self.trust_store_path = trust_store_path


class ServiceConfiguration(object):
    api_token = None  # type: str
    security = SslConfiguration  # type: Any
    connect_timeout = 10  # type: float
    read_timeout = 300  # type: float
    uris = []  # type: List[str]
    max_num_retries = 4  # type: int
    backoff_slot_size = 250  # type: int



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._lib import ConjureBeanType, ConjureUnionType, ConjureEnumType
from math import isnan, isinf
from typing import Dict, Any
import json


class ConjureEncoder(json.JSONEncoder):
    """Transforms a conjure type into json"""

    @classmethod
    def encode_conjure_bean_type(cls, obj):
        # type: (ConjureBeanType) -> Any
        """Encodes a conjure bean into json"""
        encoded = {}  # type: Dict[str, Any]
        for attribute_name, field_definition in obj._fields().items():
            encoded[field_definition.identifier] = cls.do_encode(
                getattr(obj, attribute_name)
            )
        return encoded

    @classmethod
    def encode_conjure_union_type(cls, obj):
        # type: (ConjureUnionType) -> Any
        """Encodes a conjure union into json"""
        encoded = {}  # type: Dict[str, Any]
        encoded["type"] = obj.type
        for attr, field_definition in obj._options().items():
            if field_definition.identifier == obj.type:
                attribute = attr
                break
        else:
            raise ValueError(
                "could not find attribute for union " +
                "member {0} of type {1}".format(obj.type, obj.__class__)
            )

        defined_field_definition = obj._options()[attribute]
        encoded[defined_field_definition.identifier] = cls.do_encode(
            getattr(obj, attribute)
        )
        return encoded

    @classmethod
    def encode_primitive(cls, obj):
        if isinstance(obj, float) and isnan(obj):
            return 'NaN'
        if isinstance(obj, float) and isinf(obj):
            return '{}Infinity'.format('-' if obj < 0 else '')
        return obj

    @classmethod
    def do_encode(cls, obj):
        # type: (Any) -> Any
        """Encodes the passed object into json"""
        if isinstance(obj, ConjureBeanType):
            return cls.encode_conjure_bean_type(obj)

        elif isinstance(obj, ConjureUnionType):
            return cls.encode_conjure_union_type(obj)

        elif isinstance(obj, ConjureEnumType):
            return obj.value

        elif isinstance(obj, list):
            return list(map(cls.do_encode, obj))

        elif isinstance(obj, dict):
            return {cls.do_encode(key): cls.do_encode(value)
                    for key, value in obj.items()}

        else:
            return cls.encode_primitive(obj)

    def default(self, obj):
        # type: (Any) -> Any
        return self.do_encode(obj)
